from einops import rearrange
import torch
import torch.nn as nn
import torch.nn.functional as F

class VanillaAttention(nn.Module):
    """Multi-headed attention.

    See "Attention Is All You Need" for more details.
    """

    def __init__(
        self,
        embed_dim,
        num_heads,
        causal=False,
        dropout=0.0,
    ):
        super().__init__()
        self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.causal = causal
        self.head_dim = self.embed_dim // self.num_heads
        assert (
            self.embed_dim % self.num_heads == 0
        ), "embed_dim must be divisible by num_heads"

        self.q_proj = nn.Linear(self.embed_dim, self.embed_dim)
        self.k_proj = nn.Linear(self.embed_dim, self.embed_dim)
        self.v_proj = nn.Linear(self.embed_dim, self.embed_dim)
        self.out_proj = nn.Linear(self.embed_dim, self.embed_dim)
        self.dropout = nn.Dropout(dropout)
        self.scale = self.head_dim ** -0.5

        self.reset_parameters()

    def reset_parameters(self):
        nn.init.xavier_uniform_(self.q_proj.weight)
        nn.init.normal_(self.q_proj.bias)
        nn.init.xavier_uniform_(self.k_proj.weight)
        nn.init.normal_(self.k_proj.bias)
        nn.init.xavier_uniform_(self.v_proj.weight)
        nn.init.normal_(self.v_proj.bias)
        nn.init.xavier_uniform_(self.out_proj.weight)
        nn.init.normal_(self.out_proj.bias)

    def forward(
        self,
        x,
        y=None,
        attn_mask=None,
    ):
        """
        - x: (B, N, E), where N is the sequence length, B is the batch size, E is
          the embedding dimension.
        - y: (B, M, E), where M is the sequence length, B is the batch size, E is
          the embedding dimension.
        - attn_mask: (N, M).
        """
        if y == None:
            y = x
        # B, N, E
        q = self.q_proj(x)
        # B, M, E
        k = self.k_proj(y)
        # B, M, D
        v = self.v_proj(y)
        # B, H, N, E
        q, k, v = map(lambda x: rearrange(x, 'b n (h d) -> b h n d', h=self.num_heads), [q, k, v])

        b, h, n, d = q.shape
        scaling = d ** -0.5
        score = torch.einsum('bhnd,bhmd->bhnm', q, k) * scaling
        q = q * scaling

        if self.causal:
            if attn_mask == None:
                attn_mask = (torch.triu(torch.ones(n, n))==1).transpose(0, 1)
                attn_mask = attn_mask.float().masked_fill(attn_mask==0, float('-inf')).to(q)
            score = score.masked_fill(attn_mask==float("-inf"), float("-inf"))
        weights = F.softmax(score, dim=-1)
        weights = self.dropout(weights)
        attn_output = torch.einsum('bhnm,bhmd->bhnd', weights, v)
        # reshape
        attn_output = rearrange(attn_output, 'b h n d -> b n (h d)')
        # B, N, E
        attn_output = self.out_proj(attn_output)

        return attn_output


import torch.nn as nn
import torch.nn.functional as F
from .block import TransformerBlock

class VanillaTransformer(nn.Module):
    def __init__(
        self,
        num_layers,
        embed_dim,
        hidden_dim,
        num_heads,
        causal=False,
        attn_dropout=0.0,
        ffn_dropout=0.0,
        attn_type="vanilla",
        ffn_type="vanilla",
        norm_type="ln",
        pre_norm=False,
        act_fun="gelu",
    ):
        super().__init__()
        self.layers = nn.ModuleList([])
        for _ in range(num_layers):
            self.layers.append(
                TransformerBlock(
                    embed_dim=embed_dim,
                    hidden_dim=hidden_dim,
                    num_heads=num_heads,
                    causal=causal,
                    attn_dropout=attn_dropout,
                    ffn_dropout=ffn_dropout,
                    attn_type=attn_type,
                    ffn_type=ffn_type,
                    norm_type=norm_type,
                    pre_norm=pre_norm,
                    act_fun=act_fun,
                )
            )

    def forward(self, x, y=None):
        for layer in self.layers:
            x = layer(x)

        return x


import torch.nn.functional as F
import torch.nn as nn
import torch
from .norm import *

NEG_INFINITY = float('-inf')
POS_INFINITY = float('inf')

def get_activation_fn(activation):
    if activation == "gelu":
        return F.gelu
    elif activation == "relu":
        return F.relu
    elif activation == "elu":
        return F.elu
    elif activation == "sigmoid":
        return F.sigmoid
    elif activation == "exp":
        return torch.exp
    elif activation == "leak":
        return F.leaky_relu
    elif activation == "1+elu":
        def f(x):
            return 1 + F.elu(x)
        return f
    elif activation == "silu":
        return F.silu
    else:
        return lambda x: x

def get_norm(norm_type, embed_dim):
    if norm_type == "rmsnorm":
        return RMSNorm(embed_dim)
    elif norm_type == "gatedrmsnorm":
        return GatedRMSNorm(embed_dim)
    elif norm_type == "simplermsnorm":
        return SimpleRMSNorm(embed_dim)
    elif norm_type == "scalenorm":
        return ScaleNorm(embed_dim)
    else:
        return nn.LayerNorm(embed_dim)


# https://github.com/bzhangGo/rmsnorm/blob/master/rmsnorm_torch.py
# https://github.com/bzhangGo/zero/blob/master/modules/rela.py
import torch
import torch.nn as nn
import torch.functional as F

class SimpleRMSNorm(nn.Module):
    def __init__(self, d, eps=1e-8, bias=False):
        """
            Root Mean Square Layer Normalization
        :param d: model size
        :param p: partial RMSNorm, valid value [0, 1], default -1.0 (disabled)
        :param eps:  epsilon value, default 1e-8
        :param bias: whether use bias term for RMSNorm, disabled by
            default because RMSNorm doesn't enforce re-centering invariance.
        """
        super().__init__()
        self.eps = eps
        self.d = d

    def forward(self, x):
        norm_x = x.norm(2, dim=-1, keepdim=True)
        d_x = self.d
        rms_x = norm_x * d_x ** (-1. / 2)
        x_normed = x / (rms_x + self.eps)

        return x_normed

class RMSNorm(nn.Module):
    def __init__(self, d, p=-1., eps=1e-8, bias=False):
        """
            Root Mean Square Layer Normalization
        :param d: model size
        :param p: partial RMSNorm, valid value [0, 1], default -1.0 (disabled)
        :param eps:  epsilon value, default 1e-8
        :param bias: whether use bias term for RMSNorm, disabled by
            default because RMSNorm doesn't enforce re-centering invariance.
        """
        super().__init__()

        self.eps = eps
        self.d = d
        self.p = p
        self.bias = bias

        self.scale = nn.Parameter(torch.ones(d))
        self.register_parameter("scale", self.scale)

        if self.bias:
            self.offset = nn.Parameter(torch.zeros(d))
            self.register_parameter("offset", self.offset)

    def forward(self, x):
        if self.p < 0. or self.p > 1.:
            norm_x = x.norm(2, dim=-1, keepdim=True)
            d_x = self.d
        else:
            partial_size = int(self.d * self.p)
            partial_x, _ = torch.split(x, [partial_size, self.d - partial_size], dim=-1)

            norm_x = partial_x.norm(2, dim=-1, keepdim=True)
            d_x = partial_size

        rms_x = norm_x * d_x ** (-1. / 2)
        x_normed = x / (rms_x + self.eps)

        if self.bias:
            return self.scale * x_normed + self.offset

        return self.scale * x_normed

class GatedRMSNorm(nn.Module):
    def __init__(self, d, eps=1e-8, bias=False):
        """
            Root Mean Square Layer Normalization
        :param d: model size
        :param eps:  epsilon value, default 1e-8
        :param bias: whether use bias term for RMSNorm, disabled by
            default because RMSNorm doesn't enforce re-centering invariance.
        """
        super().__init__()

        self.eps = eps
        self.d = d
        self.bias = bias

        self.scale = nn.Parameter(torch.ones(d))
        self.register_parameter("scale", self.scale)
        self.gate = nn.Parameter(torch.ones(d))
        self.register_parameter("scale", self.scale)


    def forward(self, x):
        norm_x = x.norm(2, dim=-1, keepdim=True)
        d_x = self.d

        rms_x = norm_x * d_x ** (-1. / 2)
        x_normed = x / (rms_x + self.eps)

        return self.scale * x_normed * torch.sigmoid(self.gate * x)


import torch.nn as nn
import torch.nn.functional as F
from ..attention import VanillaAttention
from ..ffn import VanillaFeedForward
from ..utils import get_norm

class TransformerBlock(nn.Module):
    def __init__(
        self,
        embed_dim,
        hidden_dim,
        num_heads,
        causal=False,
        attn_dropout=0.0,
        ffn_dropout=0.0,
        attn_type="vanilla",
        ffn_type="vanilla",
        norm_type="ln",
        pre_norm=False,
        act_fun="gelu",
    ):
        super().__init__()
        if attn_type == "vanilla":
            self.attn = VanillaAttention(embed_dim=embed_dim, num_heads=num_heads, causal=causal, dropout=attn_dropout)
        else:
            self.attn = VanillaAttention(embed_dim=embed_dim, num_heads=num_heads, causal=causal, dropout=attn_dropout)

        if ffn_type == "vanilla":
            self.ffn = VanillaFeedForward(embed_dim=embed_dim, hidden_dim=hidden_dim, act_dropout=ffn_dropout, act_fun=act_fun)
        else:
            self.ffn = VanillaFeedForward(embed_dim=embed_dim, hidden_dim=hidden_dim, act_dropout=ffn_dropout, act_fun=act_fun)

        self.attn_norm = get_norm(norm_type, embed_dim)
        self.ffn_norm = get_norm(norm_type, embed_dim)

        self.pre_norm = pre_norm
        if self.pre_norm:
            self.forward = self.forward_pre_norm
        else:
            self.forward = self.forward_post_norm

    def forward_pre_norm(self, x, y=None):
        # attention
        x = self.attn_norm(x)
        if y == None:
            y = x
        x = x + self.attn(x, y, y)
        # ffn
        x = self.ffn_norm(x)
        x = x + self.ffn(x)

        return x

    def forward_post_norm(self, x, y=None):
        # attention
        if y == None:
            y = x
        x = x + self.attn(x, y, y)
        x = self.attn_norm(x)
        # ffn
        x = x + self.ffn(x)
        x = self.ffn_norm(x)

        return x


import torch.nn as nn
import torch

class OffsetScale(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.gamma = nn.Parameter(torch.ones(dim))
        self.beta = nn.Parameter(torch.zeros(dim))
        nn.init.normal_(self.gamma, std=0.02)

    def forward(self, x):
        out = torch.einsum('... d, d -> ... d', x, self.gamma) + self.beta

        return out


import torch.nn as nn
from ..utils import get_activation_fn

class VanillaFeedForward(nn.Module):
    def __init__(
        self,
        embed_dim,
        hidden_dim,
        act_dropout=0.0,
        final_dropout=0.0,
        activation="relu",
    ):
        super().__init__()
        self.embed_dim = embed_dim
        self.hidden_dim = hidden_dim
        self.fc1 = nn.Linear(embed_dim, hidden_dim)
        self.act_fun = get_activation_fn(activation)
        self.act_dropout = nn.Dropout(act_dropout)
        self.fc2 = nn.Linear(hidden_dim, embed_dim)
        self.final_dropout = nn.Dropout(final_dropout)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act_fun(x)
        x = self.act_dropout(x)
        x = self.fc2(x)
        x = self.final_dropout(x)

        return x


# for flash
import torch
from torch import nn

class ScaleNorm(nn.Module):
    def __init__(self, d, eps=1e-5):
        super().__init__()
        self.eps = eps
        self.d = d
        self.scala = nn.Parameter(torch.ones(1))

    def forward(self, x):
        norm_x = x.norm(2, dim=-1, keepdim=True)
        d_x = self.d
        rms_x = norm_x * d_x ** (-1. / 2)
        x = x * self.scala / (rms_x + self.eps)

        return x


from setuptools import setup, find_packages

setup(
    name='tacos',
    version='0.0.1',
    description='Transformer components',
    author='Doraemonzzz',
    author_email='doraemon_zzz@163.com',
    url='https://github.com/Doraemonzzz/transformer-components',
    install_requires=[
        'torch',
        'einops',
    ],
    packages=find_packages(
        exclude=[
                "examples",
                "examples.*",
                "data",
                "data.*",
        ]
    ),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
    ],

)